import { NextApiRequest, NextApiResponse } from "next"
import logging, { Logging } from "../../lib/backend/logging"
import getPlayerByNameDB from "../../lib/backend/components/getPlayerByNameDB"
import checkPasswordIsValid from "../../lib/backend/components/checkPasswordIsValid"
import createTokenDB from "../../lib/backend/components/createTokenDB"
import sendResponse from "../../lib/backend/components/sendResponse"
import { setCookie } from "cookies-next"
import { Player, Token } from "@prisma/client"
import prisma from "../../lib/prisma"
import jwt from "jsonwebtoken"
import errors from "../../lib/backend/errors"
import sendError from "../../lib/backend/components/sendError"
import getUserFromBody from "../../lib/backend/components/getUserFromBody"

interface Data {
  loggedIn: boolean
}

export default async function login(
  req: NextApiRequest,
  res: NextApiResponse<any>
) {
  return preCheck(req, res, () =>
    getUserFromBody(req, res, (username, password) =>
      getPlayerByNameDB(req, res, username, (player) => {
        checkPasswordIsValid(req, res, player, password, () => {
          createTokenDB(player, "REFRESH", (newToken) => {
            sendResponse(req, res, loginResponse(player, newToken, req, res))
          })
        })
      })
    )
  ).catch((err) => sendError(req, res, err))
}

async function preCheck(
  req: NextApiRequest,
  res: NextApiResponse,
  next: () => void
) {
  const oldRefreshToken = req.cookies.token

  try {
    if (!oldRefreshToken) return sendError(req, res, errors.noCookie)
    // Check for old cookie, if unused invalidate it
    const tokenData = jwt.verify(
      oldRefreshToken,
      process.env.ACCESS_TOKEN_SECRET as string
    )

    if (!tokenData || typeof tokenData === "string") return

    const oldDBToken = await prisma.token.findUniqueOrThrow({
      where: {
        id: tokenData.id,
      },
    })

    if (oldDBToken.used)
      logging("Old login token was used: " + oldDBToken.id, ["debug"], req)
    else {
      await prisma.token.update({
        where: {
          id: tokenData.id,
        },
        data: {
          used: true,
        },
      })
      await logging("Old login token has been invalidated.", ["debug"], req)
    }
  } catch (err) {
    // err is expected if no correct cookie, just continue, otherwise it has been invalidated above
  }

  return next()
}

function loginResponse<T>(
  player: Player,
  newToken: [string, Token],
  req: NextApiRequest,
  res: NextApiResponse<T>
) {
  // const { player, req, res } = payload
  const [newTokenValue, newTokenDB] = newToken

  // Set login cookie
  setCookie("token", newTokenValue, {
    req,
    res,
    maxAge: 172800,
    httpOnly: true,
    sameSite: true,
    secure: true,
    path: "/api",
  })

  // deleteCookie("token", {
  //   req,
  //   res,
  //   httpOnly: true,
  //   sameSite: true,
  //   secure: true,
  //   path: "/api",
  // })

  // Successfull response
  return {
    message:
      "User " +
      player.id +
      " logged in and generated Refresh-Token: " +
      newTokenDB.id,
    body: { loggedIn: true },
    type: ["debug", "info.cyan"] as Logging[],
  }
}
